/*
 * Copyright (c) 2015-2015 Texas Instruments Incorporated
 *
 * Copyright (c) 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* @(#)log.c	8.2 (Berkeley) 11/30/93 */
#include <float.h>
#include <stdint.h>
#include <math.h>

#include "mathimpl.h"

/* Table-driven natural logarithm.
 *
 * This code was derived, with minor modifications, from:
 *	Peter Tang, "Table-Driven Implementation of the
 *	Logarithm in IEEE Floating-Point arithmetic." ACM Trans.
 *	Math Software, vol 16. no 4, pp 378-400, Dec 1990).
 *
 * Calculates log(2^m*F*(1+f/F)), |f/j| <= 1/256,
 * where F = j/128 for j an integer in [0, 128].
 *
 * log(2^m) = log2_hi*m + log2_tail*m
 * since m is an integer, the dominant term is exact.
 * m has at most 10 digits (for subnormal numbers),
 * and log2_hi has 11 trailing zero bits.
 *
 * log(F) = logF_hi[j] + logF_lo[j] is in tabular form in log_table.h
 * logF_hi[] + 512 is exact.
 *
 * log(1+f/F) = 2*f/(2*F + f) + 1/12 * (2*f/(2*F + f))**3 + ...
 * the leading term is calculated to extra precision in two
 * parts, the larger of which adds exactly to the dominant
 * m and F terms.
 * There are two cases:
 *	1. when m, j are non-zero (m | j), use absolute
 *	   precision for the leading term.
 *	2. when m = j = 0, |1-x| < 1/256, and log(x) ~= (x-1).
 *	   In this case, use a relative precision of 24 bits.
 * (This is done differently in the original paper)
 *
 * Special cases:
 *	0	return signalling -Inf
 *	neg	return signalling NaN
 *	+Inf	return +Inf
*/

#define N 128

/* Table of log(Fj) = logF_head[j] + logF_tail[j], for Fj = 1+j/128.
 * Used for generation of extend precision logarithms.
 * The constant 35184372088832 is 2^45, so the divide is exact.
 * It ensures correct reading of logF_head, even for inaccurate
 * decimal-to-binary conversion routines.  (Everybody gets the
 * right answer for integers less than 2^53.)
 * Values for log(F) were generated using error < 10^-57 absolute
 * with the bc -l package.
*/
_DATA_ACCESS static const __ieee64_t	A1 = 	  .08333333333333178827L;
_DATA_ACCESS static const __ieee64_t	A2 = 	  .01250000000377174923L;
_DATA_ACCESS static const __ieee64_t	A3 =	 .002232139987919447809L;
_DATA_ACCESS static const __ieee64_t	A4 =	.0004348877777076145742L;

_DATA_ACCESS static const __ieee64_t logF_head[N+1] = {
	0.L,
	.007782140442060381246L,
	.015504186535963526694L,
	.023167059281547608406L,
	.030771658666765233647L,
	.038318864302141264488L,
	.045809536031242714670L,
	.053244514518837604555L,
	.060624621816486978786L,
	.067950661908525944454L,
	.075223421237524235039L,
	.082443669210988446138L,
	.089612158689760690322L,
	.096729626458454731618L,
	.103796793681567578460L,
	.110814366340264314203L,
	.117783035656430001836L,
	.124703478501032805070L,
	.131576357788617315236L,
	.138402322859292326029L,
	.145182009844575077295L,
	.151916042025732167530L,
	.158605030176659056451L,
	.165249572895390883786L,
	.171850256926518341060L,
	.178407657472689606947L,
	.184922338493834104156L,
	.191394852999565046047L,
	.197825743329758552135L,
	.204215541428766300668L,
	.210564769107350002741L,
	.216873938300523150246L,
	.223143551314024080056L,
	.229374101064877322642L,
	.235566071312860003672L,
	.241719936886966024758L,
	.247836163904594286577L,
	.253915209980732470285L,
	.259957524436686071567L,
	.265963548496984003577L,
	.271933715484010463114L,
	.277868451003087102435L,
	.283768173130738432519L,
	.289633292582948342896L,
	.295464212893421063199L,
	.301261330578199704177L,
	.307025035294827830512L,
	.312755710004239517729L,
	.318453731118097493890L,
	.324119468654316733591L,
	.329753286372579168528L,
	.335355541920762334484L,
	.340926586970454081892L,
	.346466767346100823488L,
	.351976423156884266063L,
	.357455888922231679316L,
	.362905493689140712376L,
	.368325561158599157352L,
	.373716409793814818840L,
	.379078352934811846353L,
	.384411698910298582632L,
	.389716751140440464951L,
	.394993808240542421117L,
	.400243164127459749579L,
	.405465108107819105498L,
	.410659924985338875558L,
	.415827895143593195825L,
	.420969294644237379543L,
	.426084395310681429691L,
	.431173464818130014464L,
	.436236766774527495726L,
	.441274560805140936281L,
	.446287102628048160113L,
	.451274644139630254358L,
	.456237433481874177232L,
	.461175715122408291790L,
	.466089729924533457960L,
	.470979715219073113985L,
	.475845904869856894947L,
	.480688529345570714212L,
	.485507815781602403149L,
	.490303988045525329653L,
	.495077266798034543171L,
	.499827869556611403822L,
	.504556010751912253908L,
	.509261901790523552335L,
	.513945751101346104405L,
	.518607764208354637958L,
	.523248143765158602036L,
	.527867089620485785417L,
	.532464798869114019908L,
	.537041465897345915436L,
	.541597282432121573947L,
	.546132437597407260909L,
	.550647117952394182793L,
	.555141507540611200965L,
	.559615787935399566777L,
	.564070138285387656651L,
	.568504735352689749561L,
	.572919753562018740922L,
	.577315365035246941260L,
	.581691739635061821900L,
	.586049045003164792433L,
	.590387446602107957005L,
	.594707107746216934174L,
	.599008189645246602594L,
	.603290851438941899687L,
	.607555250224322662688L,
	.611801541106615331955L,
	.616029877215623855590L,
	.620240409751204424537L,
	.624433288012369303032L,
	.628608659422752680256L,
	.632766669570628437213L,
	.636907462236194987781L,
	.641031179420679109171L,
	.645137961373620782978L,
	.649227946625615004450L,
	.653301272011958644725L,
	.657358072709030238911L,
	.661398482245203922502L,
	.665422632544505177065L,
	.669430653942981734871L,
	.673422675212350441142L,
	.677398823590920073911L,
	.681359224807238206267L,
	.685304003098281100392L,
	.689233281238557538017L,
	.693147180560117703862
};

_DATA_ACCESS static const __ieee64_t logF_tail[N+1] = {
	0.L,
	-.00000000000000543229938420049L,
	 .00000000000000172745674997061L,
	-.00000000000001323017818229233L,
	-.00000000000001154527628289872L,
	-.00000000000000466529469958300L,
	 .00000000000005148849572685810L,
	-.00000000000002532168943117445L,
	-.00000000000005213620639136504L,
	-.00000000000001819506003016881L,
	 .00000000000006329065958724544L,
	 .00000000000008614512936087814L,
	-.00000000000007355770219435028L,
	 .00000000000009638067658552277L,
	 .00000000000007598636597194141L,
	 .00000000000002579999128306990L,
	-.00000000000004654729747598444L,
	-.00000000000007556920687451336L,
	 .00000000000010195735223708472L,
	-.00000000000017319034406422306L,
	-.00000000000007718001336828098L,
	 .00000000000010980754099855238L,
	-.00000000000002047235780046195L,
	-.00000000000008372091099235912L,
	 .00000000000014088127937111135L,
	 .00000000000012869017157588257L,
	 .00000000000017788850778198106L,
	 .00000000000006440856150696891L,
	 .00000000000016132822667240822L,
	-.00000000000007540916511956188L,
	-.00000000000000036507188831790L,
	 .00000000000009120937249914984L,
	 .00000000000018567570959796010L,
	-.00000000000003149265065191483L,
	-.00000000000009309459495196889L,
	 .00000000000017914338601329117L,
	-.00000000000001302979717330866L,
	 .00000000000023097385217586939L,
	 .00000000000023999540484211737L,
	 .00000000000015393776174455408L,
	-.00000000000036870428315837678L,
	 .00000000000036920375082080089L,
	-.00000000000009383417223663699L,
	 .00000000000009433398189512690L,
	 .00000000000041481318704258568L,
	-.00000000000003792316480209314L,
	 .00000000000008403156304792424L,
	-.00000000000034262934348285429L,
	 .00000000000043712191957429145L,
	-.00000000000010475750058776541L,
	-.00000000000011118671389559323L,
	 .00000000000037549577257259853L,
	 .00000000000013912841212197565L,
	 .00000000000010775743037572640L,
	 .00000000000029391859187648000L,
	-.00000000000042790509060060774L,
	 .00000000000022774076114039555L,
	 .00000000000010849569622967912L,
	-.00000000000023073801945705758L,
	 .00000000000015761203773969435L,
	 .00000000000003345710269544082L,
	-.00000000000041525158063436123L,
	 .00000000000032655698896907146L,
	-.00000000000044704265010452446L,
	 .00000000000034527647952039772L,
	-.00000000000007048962392109746L,
	 .00000000000011776978751369214L,
	-.00000000000010774341461609578L,
	 .00000000000021863343293215910L,
	 .00000000000024132639491333131L,
	 .00000000000039057462209830700L,
	-.00000000000026570679203560751L,
	 .00000000000037135141919592021L,
	-.00000000000017166921336082431L,
	-.00000000000028658285157914353L,
	-.00000000000023812542263446809L,
	 .00000000000006576659768580062L,
	-.00000000000028210143846181267L,
	 .00000000000010701931762114254L,
	 .00000000000018119346366441110L,
	 .00000000000009840465278232627L,
	-.00000000000033149150282752542L,
	-.00000000000018302857356041668L,
	-.00000000000016207400156744949L,
	 .00000000000048303314949553201L,
	-.00000000000071560553172382115L,
	 .00000000000088821239518571855L,
	-.00000000000030900580513238244L,
	-.00000000000061076551972851496L,
	 .00000000000035659969663347830L,
	 .00000000000035782396591276383L,
	-.00000000000046226087001544578L,
	 .00000000000062279762917225156L,
	 .00000000000072838947272065741L,
	 .00000000000026809646615211673L,
	-.00000000000010960825046059278L,
	 .00000000000002311949383800537L,
	-.00000000000058469058005299247L,
	-.00000000000002103748251144494L,
	-.00000000000023323182945587408L,
	-.00000000000042333694288141916L,
	-.00000000000043933937969737844L,
	 .00000000000041341647073835565L,
	 .00000000000006841763641591466L,
	 .00000000000047585534004430641L,
	 .00000000000083679678674757695L,
	-.00000000000085763734646658640L,
	 .00000000000021913281229340092L,
	-.00000000000062242842536431148L,
	-.00000000000010983594325438430L,
	 .00000000000065310431377633651L,
	-.00000000000047580199021710769L,
	-.00000000000037854251265457040L,
	 .00000000000040939233218678664L,
	 .00000000000087424383914858291L,
	 .00000000000025218188456842882L,
	-.00000000000003608131360422557L,
	-.00000000000050518555924280902L,
	 .00000000000078699403323355317L,
	-.00000000000067020876961949060L,
	 .00000000000016108575753932458L,
	 .00000000000058527188436251509L,
	-.00000000000035246757297904791L,
	-.00000000000018372084495629058L,
	 .00000000000088606689813494916L,
	 .00000000000066486268071468700L,
	 .00000000000063831615170646519L,
	 .00000000000025144230728376072L,
	-.00000000000017239444525614834
};

/*
 * Extra precision variant, returning struct {double a, b;};
 * log(x) = a+b to 63 bits, with a rounded to 26 bits.
 */
struct Double
__log__D(__ieee64_t x)
{
	int m, j;
	__ieee64_t F, f, g, q, u, v, u2;
	volatile __ieee64_t u1;
	struct Double r;

	/* Argument reduction: 1 <= g < 2; x/2^m = g;	*/
	/* y = F*(1 + f/F) for |f| <= 2^-8		*/

	m = F(ilogb)(x);
	g = F(ldexp)(x, -m);
	if (m == -1022) {
                j = F(ilogb)(g), m += j;
		g = F(ldexp)(g, -j);
	}
	j = N*(g-1) + .5;
	F = (1.0/N) * j + 1;
	f = g - F;

	g = 1/(2*F+f);
	u = 2*f*g;
	v = u*u;
	q = u*v*(A1 + v*(A2 + v*(A3 + v*A4)));
	if (m | j)
		u1 = u + 513, u1 -= 513;
	else
		u1 = u, TRUNC(u1);
	u2 = (2.0*(f - F*u1) - u1*f) * g;

	u1 += m*logF_head[N] + logF_head[j];

	u2 +=  logF_tail[j]; u2 += q;
	u2 += logF_tail[N]*m;
	r.a = u1 + u2;			/* Only difference is here */
	TRUNC(r.a);
	r.b = (u1 - r.a) + u2;
	return (r);
}
